# -*- coding: utf-8 -*-

from typing import Any, Dict, Optional, Sequence, Tuple, Union


def split(
    data: Sequence,
    val_size: float = 0.1,
    test_size: float = 0.2,
    random_state: Optional[int] = None,
    shuffle: bool = True,
    stratify: Optional[Sequence] = None,
) -> Union[Dict[str, Any], Dict[str, Tuple]]:
    # pylint: disable=import-outside-toplevel
    from sklearn.model_selection import train_test_split

    def _new(x, data=data):
        return type(data)(x)

    def _split(data, size, stratify=None):
        params = (data, stratify) if stratify is not None else (data,)
        splits = train_test_split(
            *params,
            test_size=size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify
        )

        return splits

    val_size /= 1 - test_size
    splits = _split(data, test_size, stratify=stratify)

    if stratify is None:
        train_val, test = splits
        train, val = _split(train_val, val_size, stratify=stratify)
        return {
            "train": _new(train),
            "val": _new(val),
            "test": _new(test),
        }

    train_val_x, test_x, train_val_y, test_y = splits
    train_x, val_x, train_y, val_y = _split(train_val_x, val_size, stratify=train_val_y)

    return {
        "train": (_new(train_x), _new(train_y, stratify)),
        "val": (_new(val_x), _new(val_y, stratify)),
        "test": (_new(test_x), _new(test_y, stratify)),
    }



# -*- coding: utf-8 -*-

import copy
import importlib


def reflect(module_specs):
    modules = []
    for module_spec in module_specs:
        try:
            module = importlib.import_module(module_spec)
        except ModuleNotFoundError:
            module = None

        if module is not None:
            modules += [module]

    def _constructor(*args, **kwargs):
        kwargs = copy.deepcopy(kwargs)
        class_name = kwargs.pop("class")
        cls = None
        for module in modules:
            try:
                cls = getattr(module, class_name)
                break
            except AttributeError:
                pass

        if cls is None:
            raise ValueError(
                "class name `%s` not found in modules: [%r]"
                % (class_name, module_specs)
            )

        return cls(*args, **kwargs)

    return _constructor



# -*- coding: utf-8 -*-

import contextlib
import importlib
import random
from typing import List, Optional, Union

from carton.palette import Colors


def random_state(
    seed: Optional[int] = None, low: int = 0, high: int = 2 ** 32 - 1
) -> int:
    if seed is not None:
        return seed

    return random.randint(low, high)


def set_seed(seed: int, debug: bool = False) -> None:
    def _import(module):
        try:
            module = importlib.import_module(module)
        except ModuleNotFoundError:
            return None

        return module

    random.seed(seed)

    numpy = _import("numpy")
    if numpy:
        numpy.random.seed(seed)

    torch = _import("torch")
    if torch:
        torch.manual_seed(seed)
        torch.random.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        if debug:
            torch.backends.cudnn.enabled = False
            torch.backends.cudnn.benchmark = False
            torch.backends.cudnn.deterministic = True


@contextlib.contextmanager
def seed(random_seed: int):
    state = random.getstate()
    random.seed(random_seed)
    try:
        yield
    finally:
        random.setstate(state)


def random_string(
    length: int = 10,
    chars: str = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
) -> str:
    return "".join(random.choices(chars, k=length))


def random_colors(
    colors: Optional[List[str]] = None,
    num: int = 1,
    replacement: bool = False,
) -> Union[str, List[str]]:
    if colors is None:
        colors = Colors.PRESETS

    sample_function = random.choices if replacement else random.sample
    selected = sample_function(colors, k=num)

    return selected if num > 1 else selected[0]



# -*- coding: utf-8 -*-

import datetime
import numbers
from typing import NewType, Optional, Union

import dateutil
import dateutil.parser

DateTime = NewType("DateTime", Union[str, numbers.Real, datetime.datetime])


def ensure_datetime(dt: DateTime) -> datetime.datetime:
    if isinstance(dt, datetime.datetime):
        return dt

    if isinstance(dt, str):
        return dateutil.parser.parse(dt)

    if isinstance(dt, numbers.Real):
        return datetime.datetime.fromtimestamp(dt)

    raise ValueError(f"Invalid datetime type: {type(dt)}")


def utcstr_to_datetime(
    string: str, timezone: Optional[Union[datetime.tzinfo, str]] = dateutil.tz.tzlocal()
) -> datetime.datetime:
    if isinstance(timezone, str):
        timezone = dateutil.tz.gettz(timezone)

    return (
        dateutil.parser.parse(string)
        .replace(tzinfo=dateutil.tz.tzutc())
        .astimezone(timezone)
        .replace(tzinfo=None)
    )


def timestamp13_to_datetime(timestamp: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp / 1000)


def timestamp10_to_datetime(timestamp: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp)


def timestamp_to_datetime(timestamp: float) -> datetime.datetime:
    return timestamp10_to_datetime(timestamp)


def date(offset: int = 0, return_date: bool = False) -> datetime.date:
    _date = datetime.datetime.now() + datetime.timedelta(days=offset)
    if return_date:
        return _date.date()

    return datetime.datetime.fromordinal(_date.toordinal())


def today(return_date: bool = False) -> Union[datetime.datetime, datetime.date]:
    return date(offset=0, return_date=return_date)


def yesterday(return_date: bool = False) -> Union[datetime.datetime, datetime.date]:
    return date(offset=-1, return_date=return_date)


def tomorrow(return_date: bool = False) -> Union[datetime.datetime, datetime.date]:
    return date(offset=1, return_date=return_date)



# -*- coding: utf-8 -*-

import os
import shutil
import subprocess
from typing import Any


def better_display() -> None:
    # pylint: disable=import-outside-toplevel
    import pandas as pd

    pd.set_option("display.unicode.east_asian_width", True)
    pd.set_option("display.unicode.ambiguous_as_wide", True)


def identity(x: Any) -> Any:
    return x


def git_version(dirname: str) -> str:
    if not shutil.which("git"):
        return None

    dirname = os.path.abspath(os.path.expanduser(dirname))
    if not os.path.isdir(dirname):
        dirname = os.path.dirname(dirname)

    cmd = [
        "git",
        "--no-pager",
        "log",
        "--pretty=oneline",
        "--abbrev-commit",
        "-1",
    ]
    try:
        proc = subprocess.run(
            cmd,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            cwd=dirname,
        )
    except subprocess.CalledProcessError:
        return None

    return proc.stdout.decode().strip()



# -*- coding: utf-8 -*-

import glob
import itertools
import os
from typing import Iterable, Tuple, Union


def normalize_path(p: str) -> str:
    return os.path.abspath(os.path.expanduser(p))


class _Path(str):
    def __new__(cls, value):
        return super().__new__(cls, (value))

    def __call__(self, *paths):
        paths = (self,) + tuple(paths)

        return _Path(os.path.join(*paths))

    def norm(self):
        return self.__class__(normalize_path(self))


def path(root: str) -> _Path:
    return _Path(root)


def iter_file_groups(
    dirname: str,
    exts: Union[str, Iterable[str]],
    with_key: bool = False,
    missing: str = "error",
) -> Union[
    Iterable[str], Iterable[Tuple[str, ...]], Tuple[str, Iterable[Tuple[str, ...]]]
]:
    def _format_ext(ext):
        return f'.{ext.lstrip(".")}'

    def _iter_files(dirname):
        for dirpath, _, filenames in os.walk(dirname):
            for filename in filenames:
                if os.path.splitext(filename)[1] in exts:
                    yield os.path.join(dirpath, filename)

    missings = {"error", "ignore"}
    if missing not in missings:
        raise ValueError(f"Param `missing` should be in {missings}")

    if isinstance(exts, str):
        return glob.iglob(
            os.path.join(dirname, f"**/*{_format_ext(exts)}"), recursive=True
        )

    exts = {*map(_format_ext, exts)}
    num_exts = len(exts)
    files = _iter_files(dirname)
    for key, group in itertools.groupby(
        sorted(files), key=lambda x: os.path.splitext(os.path.relpath(x, dirname))[0]
    ):
        group = sorted(group, key=lambda x: os.path.splitext(x)[1])
        if len(group) != num_exts and missing == "error":
            raise RuntimeError(f"Missing files: {key}.{exts}")

        yield (key, group) if with_key else group


def read_lines(filename: str, chunk_size: int = 64 * 1024) -> Iterable[str]:
    with open(filename, mode="r") as f:
        remaining = ""
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break

            linefeed = chunk.rfind("\n")
            if linefeed > 0:
                chunk, remaining = remaining + chunk[:linefeed], chunk[linefeed + 1 :]
            else:
                remaining = ""

            for line in chunk.split("\n"):
                yield line



# -*- coding: utf-8 -*-

import collections
import warnings
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Sequence, Union

from carton.utils import identity


def iterable(x: Any) -> bool:
    return isinstance(x, collections.abc.Iterable) and not isinstance(x, (str, bytes))


def flatten_dict(d: Mapping, sep: str = ".") -> Dict[str, Any]:
    flattened = {}

    def _flatten(d, prefix):
        for key, value in d.items():
            key = str(key)
            if sep in key:
                raise ValueError(f"`sep` in `key` ({key!r}) is not allowed")

            new_prefix = sep.join([prefix, key]) if prefix else key
            if not isinstance(value, collections.abc.Mapping):
                flattened[new_prefix] = value
            else:
                _flatten(value, new_prefix)

    _flatten(d, "")

    return flattened


def roughen_dict(d: Mapping[str, Any], sep: str = ".") -> Dict[str, Any]:
    roughen = {}
    for key, value in d.items():
        if isinstance(value, collections.abc.Mapping):
            warnings.warn(f"Nested mapping found: {key!r}")

        if sep not in key:
            roughen[key] = value
            continue

        *keys, last_key = key.split(sep)
        child = roughen
        for subkey in keys:
            child = child.setdefault(subkey, {})
        child[last_key] = value

    return roughen


def chain_get(d: Mapping, keys: Union[str, Sequence[str]], sep=".") -> Any:
    if isinstance(keys, str):
        if sep not in keys:
            return d[keys]
        keys = keys.split(sep)

    for key in keys:
        d = d[key]

    return d


def collate(
    data: Iterable[Mapping],
    keys: Optional[Sequence] = None,
    collate_fn: Union[Callable, Mapping[str, Callable]] = identity,
) -> Dict:
    def _get_collate_fn(key=None):
        if isinstance(collate_fn, collections.abc.Mapping):
            return collate_fn.get(key, identity)

        return collate_fn

    if keys is not None and (
        not isinstance(keys, collections.abc.Sequence) or isinstance(keys, str)
    ):
        keys = [keys]

    collated = collections.defaultdict(list)
    first_keys = None
    for x in data:
        if not isinstance(x, collections.abc.Mapping):
            raise TypeError(f"Not a mapping: {x}")

        if first_keys is None:
            first_keys = x.keys()
        elif first_keys != x.keys():
            raise ValueError(f"Inconsistent keys: {first_keys} != {x.keys()}")

        for key in keys or first_keys:
            collated[key] += [x[key]]

    collated = {key: _get_collate_fn(key)(value) for key, value in collated.items()}

    if len(collated) > 1:
        return collated

    return next(iter(collated.values()))
